/*
 * The UDS Collection
 * Copyright (C) 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _uds_filestream_hh
#define _uds_filestream_hh


// includes

#include <uds/fdbuf.hh>
#include <uds/fdstream.hh>
#include <string>


namespace uds
{

using namespace std;


/** file stream buffer. input and output buffer (and file positions) are
	maintained independently from each other
	\author Julian v. Bock */
class filebuf : public fdbuf
{
public:
	/** default constructor */
	filebuf();
	
	/** destructor */
	virtual
	~filebuf();

	/** sets stream offset */
	virtual pos_type 
	seekoff( off_type off, ios::seekdir dir,
			 ios::openmode mode = ios::in | ios::out );

	/** sets stream position */
	virtual pos_type 
	seekpos( pos_type pos,
			 ios::openmode mode = ios::in | ios::out );

	/** \return true if the file descriptor is valid */
	bool
	is_open() const { return File() >= 0; }

protected:
	/** performs read operation
		\param buf buffer start
		\param len max. length
		\return nuber of bytes read */
	virtual int_type
	Read( char_type* buf, int_type len );

	/** performs write operation
		\param buf buffer start
		\param len length */
	virtual int_type
	Write( char_type* buf, int_type len );

	/** \return real input offset */
	off_t
	rioff() const { return ioff + ( gptr() - egptr() ); }

	/** \return real output offset */
	off_t
	rooff() const { return ooff + pptr() - pbase(); }

private:
	/// real file offset
	off_t	foff;
	/// input offset
	off_t	ioff;
	/// output offset
	off_t	ooff;

	/** performs seek
		\param off offset
		\param dir direction
		\return real file offset relative to the beginning of the file */
	off_t
	fseek( off_t off, ios::seekdir dir );
};


/** file stream class. It provides access to the file descriptor that
	is associated with the stream.
	\author Julian v. Bock */
class FileStream : public iofdstream
{
public:
	/** extra FileStream flags */
	enum
	{
		excl	= 0x1,	///< exclusive file creation
		create	= 0x2	///< create file if it does not exist
	};

	typedef unsigned char efs_flags_t;

	/** protection flags */
	enum
	{
		usr_rwx	= 00700,	///< user has read, write, and execute permission
		usr_r	= 00400,	///< user has read permission
		usr_w	= 00200,	///< user has write permission
		usr_x	= 00100,	///< user has execute permission
		grp_rwx	= 00070,	///< group has read, write, and execute permission
		grp_r	= 00040,	///< group has read permission
		grp_w	= 00020,	///< group has write permission
		grp_x	= 00010,	///< group has execute permission
		oth_rwx	= 00007,	///< others have read, write, and execute permission
		oth_r	= 00004,	///< others have read permission
		oth_w	= 00002,	///< others have write permission
		oth_x	= 00001		///< others have execute permission
	};

	/** default constructor */
	FileStream();

	/** \param fd file descriptor
		\param mode open mode
		\param pf 'preserve file' flag */
	FileStream( int fd, ios::openmode mode = ios::in | ios::out,
				bool pf = false );

	/** \param fn file name
		\param mode open mode
		\param ef extra FileStream flags
		\param prot protection bits
		\param pf 'preserve file' flag */
	FileStream( const char* fn, ios::openmode mode = ios::in | ios::out,
				efs_flags_t ef = 0, mode_t prot = 0644, bool pf = false );

	/** \param fn file name
		\param mode open mode
		\param ef extra FileStream flags
		\param prot protection bits
		\param pf 'preserve file' flag */
	FileStream( const string& fn, ios::openmode mode = ios::in | ios::out,
				efs_flags_t ef = 0, mode_t prot = 0644, bool pf = false );

	/** destructor */
	virtual
	~FileStream();
	
	/** opens a file. The old file (if there is one) is not closed if the pf
		argument is true.
		\param fd file descriptor
		\param mode open mode
		\param pf 'preserve file' flag */
	virtual void
	Open( int fd, ios::openmode mode = ios::in | ios::out, bool pf = false );
	
	/** opens a file. The old file (if there is one) is not closed if the pf
		argument is true.
		\param fn file name
		\param mode open mode
		\param ef extra FileStream flags
		\param prot protection bits
		\param pf 'preserve file' flag */
	virtual void
	Open( const char* fn, ios::openmode mode = ios::in | ios::out,
		  efs_flags_t ef = 0, mode_t prot = 0644, bool pf = false );
	
	/** opens a file. The old file (if there is one) is not closed if the pf
		argument is true.
		\param fn file name
		\param mode open mode
		\param ef extra FileStream flags
		\param prot protection bits
		\param pf 'preserve file' flag */
	virtual void
	Open( const string& fn, ios::openmode mode = ios::in | ios::out,
		  efs_flags_t ef = 0, mode_t prot = 0644, bool pf = false )
		{ Open( fn.c_str(), mode, ef, prot, pf ); }

	/** closes the file */
	virtual void
	Close();

	/** provides file descriptor access
		\return file descriptor */
	int
	File() const { return buf.File(); }

	/** \return true if a file is opened */
	bool
	IsOpen() const { return buf.is_open(); }

private:
	/// stream buffer
	filebuf buf;
	/// 'preserve file' flag
	bool	pfile;

	/** checks an open mode for sanity
		\param mode open mode */
	void
	om_sanity_check( ios::openmode mode );
};

}

#endif // _uds_filestream_hh
