/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _uds_io_util_hh
#define _uds_io_util_hh


// includes

#include <uds/fdstream.hh>
#include <uds/stdexcpt.hh>
#include <sys/endian.h>
#include <limits>
#include <unistd.h>


namespace uds
{

using namespace std;


/** endianess types */
enum
{
	end_little = 1234,		///< little-endian
	end_big = 4321,			///< big-endian
	end_net = end_big,		///< network byte order
	end_host = BYTE_ORDER	///< host byte order
};

typedef unsigned endianess_t;


/** changes the endianess of an integral type
	\param v value */
template < class Tp >
inline void
ch_endianess( Tp& v )
{
	if ( !numeric_limits< Tp >::is_integer )
		throw invalid_arg( "cannot change endianess of non-integral "
						   "type" );

	switch ( sizeof( Tp ) )
	{
	case 1:
		break;

	case 2:
		v = bswap16( v );
		break;

	case 4:
		v = bswap32( v );
		break;

	case 8:
		v = bswap64( v );
		break;

	default:
		throw invalid_arg( "cannot change endianess for integral type "
						   "which is not 1, 2, 4, or 8 bytes long" );
	}
}


/** BRead / BWrite flags */
enum
{
	brw_wait		= 0x1,	///< wait for input
	brw_readps		= 0x2,	///< read fist part of a string that exceeds the max. string length
	brw_skip		= 0x4,	///< skip [rest of the] string that is too long
	brw_lnothrow	= 0x8	///< don't throw an exception a the string to be filled is too long
};

typedef unsigned char brw_flags_t;


template < class In, class Tp >
size_t
bread_val( In& in, Tp& v, brw_flags_t flags, endianess_t end )
{
	size_t sz = BReadRaw( in, &v, sizeof( v ), flags );

	if ( end != end_host )
		ch_endianess( v );
	
	return sz;
}

template < class Out, class Tp >
void
bwrite_val( Out& out, const Tp& v, endianess_t end )
{
	if ( end != end_host )
	{
		Tp e = v;
		ch_endianess( e );
		BWriteRaw( out, &e, sizeof( e ) );
	}
	else
		BWriteRaw( out, &v, sizeof( v ) );
}


/** exception class for string length errors. It is used by the BRead family
	of functions.
	\author Julian v. Bock */
class brw_string_len_error : public runtime_error
{
public: 
	/** \param ml max. length
		\param rl requested length
		\param br number of bytes read */
	brw_string_len_error( size_t ml, size_t rl, size_t br );

	/** destructor */
	virtual 
	~brw_string_len_error();

	/** \return error description */
	virtual string 
	What() const { return "BRead* string length error"; }

	/** \return max. string length */
	size_t
	MaxLen() const { return maxlen; }

	/** \return requested string length */
	size_t
	ReqLen() const { return rlen; }

	/** \return number of bytes read */
	size_t
	BRCount() const { return bread; }

private:
	/// max. string length
	size_t	maxlen;
	/// requested string length
	size_t	rlen;
	/// number of bytes read
	size_t	bread;
};


/** reads a line from an input stream
	\param in input stream
	\param o is filled with the line
	\param app 'append eol character' flag
	\param ign 'ignore eol character' flag
	\param eol end-of-line character */
void
ReadLine( istream& in, string& o, bool app = false, bool ign = true,
		  char eol = '\n' );

/** reads a line from an input stream
	\param in input stream
	\param app 'append eol character' flag
	\param ign 'ignore eol character' flag
	\param eol end-of-line character
	\return line */
string
ReadLine( istream& in, bool app = false, bool ign = true, char eol = '\n' );

/** reads data from an input stream. If the 'brw_wait' flag is set, the function
	does not return until len bytes have been read.
	\param in input stream
	\param buf buffer
	\param len length
	\param flags brw flags. Only the brw_wait flag is recognized.
	\return number of bytes read */
size_t
BReadRaw( ifdstream& in, void* buf, size_t len, brw_flags_t flags = brw_wait );

/** reads the binary representation of an object from a stream. If the
	'brw_wait' flag is set, the function
	does not return until len bytes have been read.
	\param in input stream
	\param v object
	\param flags brw flags. Only the brw_wait flag is recognized.
	\param end endianess
	\return number of bytes read */
template < class Tp >
inline size_t
BRead( ifdstream& in, Tp& v, brw_flags_t flags = brw_wait,
	   endianess_t end = end_host )
{
	return bread_val( in, v, flags, end );
}

/** reads a binary representation of a string from a stream. The binary string
	representation must consist of the string length (size_t) and the
	string data. If the 'brw_wait' flag is set, the function
	does not return until len bytes have been read.
	\param in input stream
	\param v string
	\param flags brw flags. Note that a brw_string_len_error exception is
	thrown regardless of the brw_lnothrow flag if not all characters could be
	read.
	\param end endianess. This options affects only the length parameter.
	\param ml max. string length. If the string to be read is longer than ml
	chars, the string is not read and an exception is thrown unless the
	brw_lnothrow is set. If the brw_readps flag is set, up to ml bytes are
	read. If the brw_skip flag is set, the rest of the string is ignored.
	Note that the length of the string is not put back on stream.
	However, the brw_string_len exception contains the string length.
	\return number of bytes read */
size_t
BRead( ifdstream& in, string& v, brw_flags_t flags = brw_wait,
	   endianess_t end = end_host, size_t ml = string::npos );

/** reads data from a file descriptor. If the brw_wait flag is set,
	the function does not return until len bytes have been read.
	\param in file descriptor
	\param buf buffer
	\param len length
	\param flags brw flags
	\return number of bytes read */
size_t
BReadRaw( int in, void* buf, size_t len, brw_flags_t flags = brw_wait );

/** reads the binary representation of an object from a file descriptor.
	If the brw_wait flag is set, the function does not return until len bytes
	have been read.
	\param in file descriptor
	\param v object
	\param flags brw flags
	\param end endianess
	\return number of bytes read */
template < class Tp >
inline size_t
BRead( int in, Tp& v, brw_flags_t flags = brw_wait,
	   endianess_t end = end_host )
{
	return bread_val( in, v, flags, end );
}

/** reads a binary representation of a string from a file descriptor.
	The binary string representation must consist of the string length
	(size_t) and the string data. If the 'brw_wait' flag is set, the
	function does not return until len bytes have been read.
	\param in file descriptor
	\param v string
	\param flags brw flags
	\param end endianess. This options affects only the length parameter.
	\param ml max. string length. If the string to be read is longer than ml
	chars, the string is not read and an exception is thrown unless the
	brw_lnothrow is set. If the brw_readps flag is set, up to ml bytes are
	read. If the brw_skip flag is set, the rest of the string is ignored.
	Note that the length of the string is not put back on stream.
	However, the brw_string_len exception contains the string length.
	\return number of bytes read */
size_t
BRead( int in, string& v, brw_flags_t flags = brw_wait,
	   endianess_t end = end_host, size_t ml = string::npos );

/** writes data to an output stream.
	\param out output stream
	\param buf buffer
	\param len length */
inline void
BWriteRaw( ostream& out, const void* buf, size_t len )
{
	out.write( static_cast< const char* >( buf ), len );
}

/** writes the binary representation of an object to a stream
	\param out output stream
	\param v object
	\param end endianess */
template < class Tp >
inline void
BWrite( ostream& out, const Tp& v, endianess_t end = end_host )
{
	bwrite_val( out, v, end );
}

/** writes the binary representation of a string to a stream
	\param out output stream
	\param v object
	\param end endianess */
void
BWrite( ostream& out, const string& v, endianess_t end = end_host );

/** writes data to a file.
	\param out output stream
	\param buf buffer
	\param len length
	\return number of bytes written */
inline size_t
BWriteRaw( int out, const void* buf, size_t len )
{
	return CkStdCall( write( out, buf, len ) );
}

/** writes the binary representation of an object to a stream
	\param out file descriptor
	\param v object
	\param end endianess */
template < class Tp >
inline void
BWrite( int out, const Tp& v, endianess_t end )
{
	bwrite_val( out, v, end );
}

/** writes the binary representation of a string to a stream
	\param out file descriptor
	\param v object
	\param end endianess */
void
BWrite( int out, const string& v, endianess_t end );

}

#endif // _uds_io_util_hh
