/*
 * The UDS Collection
 * Copyright (C) 1999, 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _uds_sentinel_hh
#define _uds_sentinel_hh


// includes

#include <uds/thread.hh>
#include <map>
#include <memory>
#include <set>
#include <string>


namespace uds
{

using namespace std;


/** a malloc allocator which is used by the sentinel class. some of the
	code has been taken from the standard allocator of the GNU libstdc++. */
template < class Tp >
class malloc_alloc
{
public:
	typedef size_t		size_type;
	typedef ptrdiff_t	difference_type;
	typedef Tp*			pointer;
	typedef const Tp*	const_pointer;
	typedef Tp&			reference;
	typedef const Tp&	const_reference;
	typedef Tp			value_type;

	template < class Tp1 >
	struct rebind
	{
		typedef malloc_alloc< Tp1 > other;
	};

	malloc_alloc() throw() {}
	malloc_alloc( const malloc_alloc& ) throw() {}
	
	template < class Tp1 >
	malloc_alloc( const malloc_alloc<Tp1>& ) throw() {}
	
	~malloc_alloc() throw() {}

	pointer
	address( reference x ) { return &x; }

	const_pointer
	address( const_reference x ) const { return &x; }
	
	Tp*
	allocate( size_type n, const void* = 0 )
		{ return n ? static_cast< Tp* >( malloc( n * sizeof( Tp ) )) : 0; }

	void
	deallocate( pointer p, size_type n ) { free( p ); }

	size_type
	max_size() const throw() { return size_t(-1) / sizeof(Tp); }

	void
	construct( pointer p, const Tp& val ) { new ( p ) Tp( val ); }
	
	void
	destroy( pointer p ) { p->~Tp(); }
};



/** logs memory allocations, checks for memory leaks end creates zombie
	objects.
	\author Julian v. Bock */
class Sentinel
{
public:
	/** allocates mem
		\return new memory address
		\arg throws bad_alloc */
	void*
	Allocate( size_t size );
	
	/** releases allocated memory
		\param p previously allocated mem
		\arg throws uds::zombie_object */
	void
	Release( void* p );
	
	/** checks whether memory at a given address was previously allocated
		\param p address
		\return true if memory was allocated */
	bool
	Allocated( const void* p ) const;
	
	/** performs the check for memory leaks */
	void
	Check();

	/** registers a number of expected memory leaks
		\param l number of leaks */
	void
	Expect( size_t l ) { eleaks.insert( l ); }

	/** ignore all memory leaks from source files with a certain substring
		\param s substring */
	void
	Ignore( const string& s ) { blist.insert( s ); }

	/** ignore all memory leaks from unknown sources */
	void
	IgnoreUnknown() { ignu = true; }
	
private:
	class mem_t
	{
	public:
		/// function that allocated the memory block
		const void*	caller;
		/// size of block
		const int	size;

		/** \param c caller
			\param s size */
		mem_t( const void* c, int s ) :
			caller( c ), size( s ) {}
	};

	typedef map< const void*, mem_t > std_heap_map_t;
	typedef map< const void*, mem_t, less< const void* >,
		malloc_alloc< const pair< const void*, mem_t > > > heap_map_t;
	typedef set< size_t, less< size_t >, malloc_alloc< size_t > > el_set_t;
	typedef set< string, less< string >, malloc_alloc< string > > bl_set_t;
	
	/// holds allocated memory addresses / owners
	heap_map_t	heap_map;
	/// number[s] of expected memory leaks
	el_set_t	eleaks;
	/// ignored source files (substrings)
	bl_set_t	blist;
	/// 'ingore unknown' flag
	bool		ignu;
	/// mutex
	Mutex		mut;
	
	/** default constructor */
	Sentinel();
	
	/** destructor */
	~Sentinel();
	
	friend Sentinel& sentinel();
};


/** \return the one and only Sentinel instance */
inline Sentinel&
sentinel()
{
	static char _s[sizeof( Sentinel )];
	static Sentinel* s = new ( _s ) Sentinel();
	return *s;
}

}

#endif // _uds_sentinel_hh
