/*
 * The UDS Collection
 * Copyright (C) 2000, 2001, 2002  Julian v. Bock
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *	(See file COPYING for more information)
 */

#ifndef _uds_sys_util_hh
#define _uds_sys_util_hh


// includes

#include <sys/types.h>
#include <uds/stdexcpt.hh>


namespace uds
{

using namespace std;


/// holds the default prefix for temporary file names
extern string TmpFileName_Prefix;


/** fd_set wrapper class
	\author Julian v. Bock */
class FD_Set
{
public:
	/** default constructor */
	FD_Set() { Zero(); }
	
	/** destructor */
	~FD_Set() {}
	
	/** sets a file descriptor
		\param fd file descriptor */
	void
	Set( int fd );
	
	/** clears a file descriptor. WARNING: n (the highest numbered file
		descriptor in the set plus 1) becomes undefined if you use this
		method!
		\param fd file descriptor */
	void
	Clear( int fd );
	
	/** clears all file descriptors */
	void
	Zero();
	
	/** checks whether a file descriptor is set
		\param fd file descriptor */
	bool
	IsSet( int fd ) const;
	
	/** \return the highest numbered file descriptor in the set plus 1 */
	int
	N() const { return n; }
	
	fd_set*
	operator () () { return &fds; }
	
private:
	/// file descriptor set
	fd_set	fds;
	/// highest numbered file descriptor in the set plus 1
	int		n;
};


/** checks whether a file exists
	\param f file name
	\return true if the file exists */
bool
FileExist( const char* name );

/** checks whether a file can be accessed
	\param f file name
	\param m file mode (see man 2 access for available flags)
	\return true if the file can be accessed */
bool
FileAccess( const char* f, int m );

/** \return current working directory path */
string
CurrentWD();

/** creates a unique temporary file name. For security reasons, a file with
	the temporary name and 0600 permissions is created (and closed).
	\param pre file name prefix
	\return temporary file name */
string
TmpFileName( const string& pre );

/** creates a unique temporary file name. For security reasons, a file with
	the temporary name and 0600 permissions is created (and closed).
	\return temporary file name */
string
TmpFileName();

/** creates a unique temporary file with permission 0600.
	\param pre file name prefix
	\return file descriptor of the temporary file */
int
TmpFile( const string& pre );

/** creates a unique temporary file with permission 0600.
	\return file descriptor of the temporary file */
int
TmpFile();

/** forks the process. If the call to fork() fails, a fork_error exception
	is thrown
	\return the PID of the child process is returned in the parent's thread
	of execution, and a 0 is returned in the child's thread of
	execution.
	\arg throws fork_error */
int
Fork();

/** suspends the execution until a child process exited. If the child process
	exited abnormally, a child_error exception is thrown. The child_exception
	contains a ChildDiag diagnose object. If the call to
	waitpid fails, a system_error exception is thrown.
	\param pid the same as the pid option of the waitpid function
	\param iv if this value is returned by the child process, no exception is
	thrown. Exit( iv ) is called instead.
	\param opt wait() options. see man 2 wait.
	\return pid of the child process that exited
	\arg throws child_error, system_error */
int
Wait( int pid = -1, int iv = -1, int opt = 0 );

/** tries to create a lockfile. If the lockfile does not exist, it can be
	created, but nothing is written into it. If the file does already exist
	but is not locked, the content is left unchanged. Use WriteLockFile() to
	truncate the file and write the pid into the file.
	\param lockfile lockfile name
	\param create 'create file' flag
	\param lpid if this pointer is nonzero and the file is already locked, the
	variable pointed to by lpid id filled with the pid that is read from
	the lockfile. If the file is locked but does not contain a valid pid lpid
	is set to 0. Otherwise the value pointed to by lpid will be set to -1.
	\return lockfile file descriptor. You can use it to write to the lockfile,
	and unlock / close it. If the lock can not be engaged, -1 is returned. In
	this case you might want to check errno or throw a system_error. */
int
MkLockFile( const char* lockfile, bool create = true, pid_t* lpid = 0 );

/** truncates a file and writes the pid + newline into it. This is usually
	used for lockfiles
	\param lfd lockfile file descriptor */
void
WriteLockFile( int lfd );

/** executes a command as system() does. However, interrupts are
	not ignored. If the child process does not exit
	successfully, a child_excetpion is thrown.
	\param cmd command to execute */
void
System( const char* cmd );

/** converts a group id to the group name. If the group does not exist,
	a runtime_error is thrown.
	\param gid group id
	\return group name */
string
GroupName( gid_t gid );

/** converts a group name to the group id. If the group does not exist,
	a runtime_error is thrown.
	\param name group name
	\return group id */
gid_t
GroupID( const char* name );

/** converts a user id to the user name. If the user does not exist,
	a runtime_error is thrown.
	\param uid user id
	\return user name */
string
UserName( uid_t uid );

/** converts a user name to the user id. If the user does not exist,
	a runtime_error is thrown.
	\param name user name
	\return user id */
uid_t
UserID( const char* name );

}

#endif // _uds_sys_util_hh
